/* chilmerge.c -- merge coherent and incoherent processing NetCDF outputs in a NetCDF file */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <netcdf.h>

#include "nctools.h"
#include "chil.h"

static int ncstatus;
int ncid2;


#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); nc_close(ncid2); exit(-1); \
} 
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
void
put_char_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  short value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type short\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}
static
float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type float\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_float(ncid, varid, name, &value));
  return value;
}

static
void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] inchoherent_data.nc coherent_data.nc\n", argv[0]);
  fprintf(stderr, "Options:\n"
          "         -user s  name and email address of the user\n");
  fprintf(stderr,
	  "Comments: The \"cohproc\" processing system produces two forms of data:\n"
          "    * Coherent: Doppler spectra are analysed and Z, v and width are calculated\n"
	  "    * Incoherent: Z alone is calculated simply as the average returned power\n"
	  "   There is a problem with the coherent algorithm in that it tends to underestimate\n"
	  "   Z (by up to 10 dB) at low signal-to-noise ratios.  This program takes the Z from\n"
	  "   a NetCDF file created from incoherent, and inserts it into a NetCDF file of\n"
	  "   coherent parameters, overwriting the coherent Z.\n");
}

int
main(int argc, char **argv)
{
  char *user = NULL;
  int iarg, ncid;
  size_t ngates, nrays, ngates2, nrays2;
  size_t start[2] = {0, 0};
  size_t start2[2] = {0, 0};
  size_t count[2] = {1, 1};
  int timedimid, rangedimid, timedimid2, rangedimid2;
  int timeid, timeid2;
  int Zid, Zid2;
  float cal_Zh;
  int ray, ray2;
  int status = 0;
  float *Z = NULL;
  float *Z2 = NULL;
  float missing_value;
  float missing_value2;
  char doit = 1;
  char *Zname;
  char *comment_Zh; 
  char *str = NULL;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
        if (user) {
          fprintf(stderr, "Warning: user specified more than once:"
                  " `%s' not used\n", user);
        }
        user = argv[++iarg];
      }
      else {
        fprintf(stderr, "Error: user not specified\n");
        exit(1);
      }
    }
    ++iarg;
  }
  if (iarg >= argc) {
    Usage(argv);
    exit(0);
  }

  /* Open file and get dimension sizes */
  /* This is the incoherent data */
  handle_error(nc_open(argv[iarg], 0, &ncid));
  handle_error(nc_inq_dimid(ncid, "time", &timedimid));
  handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));
  handle_error(nc_inq_dimid(ncid, "range", &rangedimid));
  handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));
  handle_error(nc_inq_varid(ncid, "time", &timeid));

  /* This is the coherent data */
  handle_error(nc_open(argv[iarg+1], NC_WRITE | NC_NOCLOBBER, &ncid2));
  handle_error(nc_redef(ncid2));
  handle_error(nc_inq_dimid(ncid2, "time", &timedimid2));
  handle_error(nc_inq_dimlen(ncid2, timedimid2, &nrays2));
  handle_error(nc_inq_dimid(ncid2, "range", &rangedimid2));
  handle_error(nc_inq_dimlen(ncid2, rangedimid2, &ngates2));
  handle_error(nc_inq_varid(ncid2, "time", &timeid2));

  /* First see if the variable "Z" is present, then "Zh" in the files*/
  Zname = "Z";
  if (nc_inq_varid(ncid, Zname, &Zid) != NC_NOERR) {
    Zname = "Zh";
    if (nc_inq_varid(ncid, Zname, &Zid) != NC_NOERR) {
      fprintf(stderr, "Warning: Z/Zh not found in %s\n", argv[iarg]);
      doit = 0;
    }
  }
  Zname = "Z";
  if (nc_inq_varid(ncid2, Zname, &Zid2) != NC_NOERR) {
    Zname = "Zh";
    if (nc_inq_varid(ncid2, Zname, &Zid2) != NC_NOERR) {
      fprintf(stderr, "Warning: Z/Zh not found in %s\n", argv[iarg+1]);
      doit = 0;
    }
  }
  if (nrays < nrays2)  {
    fprintf(stderr, "Error: %s has more rays than %s\n", argv[iarg],argv[iarg+1]);
    exit(1);
  }


  if (doit) {
    /* Exchange Attributes */
    comment_Zh= get_char_attribute(ncid, Zid, "comment");
    cal_Zh= get_float_attribute(ncid, Zid, "calibration_applied");
    handle_error(nc_put_att_float(ncid2, Zid2, "calibration_applied",
				  NC_FLOAT, 1 , &cal_Zh));
    handle_error(nc_put_att_text(ncid2, Zid2, "comment",
				 strlen(comment_Zh), comment_Zh));
    handle_error(nct_append_attribute(ncid2, Zid2,"comment",
	      "Note that this parameter was calculated using the incoherent processing system, "
	      "while the others were calculated using coherent processing.", "\n"));
    /* Get missing values */
    missing_value = get_float_attribute(ncid, Zid, "missing_value");
    missing_value2 = get_float_attribute(ncid2, Zid2, "missing_value");

    /* Now the history and command line */
    handle_error(nct_add_history(ncid2, "Inserted incoherent reflectivity factor", user));

    str = get_char_attribute(ncid, NC_GLOBAL, "history");
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "history", "(Incoherent processing history:", "\n"));
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "history", str, " "));
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "history", ")", ""));

    handle_error(nct_add_command_line(ncid2, argc, argv));
    str = get_char_attribute(ncid, NC_GLOBAL, "command_line");
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "command_line", "(Incoherent processing command line:", "\n"));
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "command_line", str, " "));
    handle_error(nct_append_attribute(ncid2, NC_GLOBAL, "command_line", ")", ""));

    /* Out of define mode */
    handle_error(nc_enddef(ncid2));
    Z = malloc(ngates * sizeof(float));
    Z2 = malloc(ngates2 * sizeof(float));
    if (!Z || !Z2) {
      fprintf(stderr, "Out of memory\n");
      exit(2);
    }

    /* Number of gates may be different due to different processing of
       coherent and incoherent systems, although they should start at
       the same point - we leave untouched upper rays that are only in
       one system */
    if (ngates < ngates2) {
      count[1] = ngates;
    }
    else {
      count[1] = ngates2;
    }

    /* Loop through the rays */
    ray = 0;
    for (ray2 = 0; ray2 < nrays2; ray2++, ray++) {
      int j;
      float time, time2;
      start[0] = ray;
      start2[0] = ray2;
      nc_get_var1_float(ncid, timeid, start, &time); 
      nc_get_var1_float(ncid2, timeid2, start2, &time2);
      if (time2-time > 0.0001) /* less than 0.5 seconds */ {
	status = 1;
	/* Coherent (dest) has missing rays compared to incoherent (src) */
	/* Skip over some incoherent rays */
	while (ray < nrays) {
	  ray++;
	  start[0] = ray;
	  nc_get_var1_float(ncid, timeid, start, &time); 
	  if (time2-time <= 0.0001 && time2-time >= -0.0001) {
	    /* Found a matching ray */
	    status = 0;
	    break;
	  }
	  else if (time2-time < -0.0001) {
	    /* Incoherent has missing rays compared to incoherent */
	    fprintf(stderr, "Error: %s has rays that %s does not (and vice versa)\n",
		    argv[iarg+1], argv[iarg]);
	    status = 1;
	    break;
	  }
	}
	if (status && ray >= nrays) {
	  fprintf(stderr, "Error: %s ended abruptly\n", argv[iarg]);
	}
      }
      else if (time2-time < -0.0001) /* less than 0.5 seconds */ {
	/* Incoherent has missing rays compared to incoherent */
	/* Quit with an error */
	fprintf(stderr, "Error: %s has rays that %s does not (%f - %f < -0.0001)\n",
		argv[iarg+1], argv[iarg], time2, time);
	status = 1;
	break;
      }

      if (!status) {
	handle_error(nc_get_vara_float(ncid, Zid, start, count, Z));
	handle_error(nc_get_vara_float(ncid2, Zid2, start2, count, Z2));
	/* Only copy data over if both systems see cloud - this removes
	   a nasty bleeding problem with the incoherent.  Also use the
	   missing_value of the coherent reflectivity */
	for (j = 0; j < count[1]; j++) {
	  if (Z[j] == missing_value || Z2[j] == missing_value2) {
	    Z[j] = missing_value2;
	  }
	}
	handle_error(nc_put_vara_float(ncid2, Zid2, start2, count, Z));
      }
    }
    free(Z);
    free(Z2);
  }
  nc_close(ncid);
  nc_close(ncid2);
  exit(status);
}









